﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.ComponentModel;

namespace AsyncExecution
{
    /// <summary>
    /// <para>This class provides the possibility to run methods that look synchronously in an asynchronous fashion.</para>       
    /// <para>For this a method has to have a return type of IEnumerator&lt;SwitchExecutionContext&gt;</para>
    /// <para>Whenever a "yield return SwitchExecutionContext.ToUIThread" is done the execution switches to the UI thread.</para>
    /// <para>Whereas when a "yield return SwitchExecutionContext.ToWorkerThread" is done the execution switches to the UI thread.</para>
    /// <para>Such a method is started using the StartExecution method(s)</para>    
    /// <para>For Debugging purposes is it possible to make the AsyncExecutor execute everything on the UI thread by setting <see cref="SyncMode"/> to true</para>
    /// </summary>
    public class AsyncExecutor : INotifyPropertyChanged
    {
        /// <summary>
        /// <para>Initializes a new instance of the <see cref="AsyncExecutor"/> class.</para>
        /// <para>Has to be called on the UI thread!</para>
        /// </summary>
        public AsyncExecutor()
        {
            uiSyncContext = SynchronizationContext.Current;

            if (uiSyncContext == null)
                throw new ApplicationException("SynchronizationContext is null. Please make sure to call the ctor from the UI thread!");
        }

        private SynchronizationContext uiSyncContext = null;

        /// <summary>
        /// Gets or sets a value indicating whether the Async Executor should run synchronously. This is helpful for debugging.
        /// </summary>
        /// <value>Current Sync Mode</value>
        public bool SyncMode { get; set; }

        /// <summary>
        /// Starts the execution.
        /// </summary>
        /// <typeparam name="InputType">The type of the parameter</typeparam>
        /// <param name="methodToExecute">The method to execute.</param>
        /// <param name="parameter">The parameter.</param>
        [System.Diagnostics.DebuggerStepThrough]
        public void StartExecution<InputType>(ExecutionDelegate<InputType> methodToExecute, InputType parameter)
        {
            if (SyncMode)
            {
                Execute<InputType>(methodToExecute, parameter);
            }
            else
            {
                NoParametersDelegate runIt = delegate
                {
                    IsBusy = true;
                    Execute<InputType>(methodToExecute, parameter);
                    IsBusy = false;
                };

                runIt.BeginInvoke(delegate(IAsyncResult iar) { runIt.EndInvoke(iar); }, null);
            }
        }

        /// <summary>
        /// <para>Executes the specified method using the current thread as background thread.</para>
        /// <para>This can be used to call methods synchronously which give back an IEnumerator&lt;SwitchExecutionContext&gt; from 
        /// a method started using one of the StartExecution overloads.</para>
        /// <para>You have to switch the current context to the Aync thread before calling this method.</para>
        /// </summary>
        /// <typeparam name="InputType">The type of the parameter</typeparam>
        /// <param name="methodToExecute">The method to execute.</param>
        /// <param name="parameter">The parameter.</param>
        [System.Diagnostics.DebuggerStepThrough]
        public void Execute<InputType>(ExecutionDelegate<InputType> methodToExecute, InputType parameter)
        {
            IEnumerator<SwitchExecutionContext> enumerator = methodToExecute(parameter);

            if (SyncMode || (SynchronizationContext.Current != null && SynchronizationContext.Current.Equals(uiSyncContext)))          
            {
                while (enumerator.MoveNext())
                {
                }
            }
            else
            {
                bool moveNext = true;
                SwitchExecutionContext mode = SwitchExecutionContext.ToUIThread;
                while (moveNext)
                {
                    if (mode == SwitchExecutionContext.ToUIThread)
                    {
                        //On UI thread
                        uiSyncContext.Send(delegate
                        {
                            moveNext = enumerator.MoveNext();
                            if (moveNext)
                            {
                                mode = enumerator.Current;
                            }
                        }, null);
                    }
                    else
                    {
                        //On the current (async) thread
                        moveNext = enumerator.MoveNext();
                        if (moveNext)
                        {
                            mode = enumerator.Current;
                        }
                    }
                }
            }
        }




        /// <summary>
        /// Starts the execution on a worker thread.
        /// </summary>
        /// <param name="methodToExecute">The method to execute.</param>
        [System.Diagnostics.DebuggerStepThrough]
        public void StartExecution(ExecutionDelegateNoParameter methodToExecute)
        {
            if (SyncMode)
            {
                Execute(methodToExecute);
            }
            else
            {

                SynchronizationContext uiSyncContext = SynchronizationContext.Current;

                NoParametersDelegate runIt = delegate
                {
                    IsBusy = true;
                    Execute(methodToExecute);
                    IsBusy = false;
                };

                runIt.BeginInvoke(delegate(IAsyncResult iar) { runIt.EndInvoke(iar); }, null);
            }
        }



        /// <summary>
        /// Executes the specified method using the current thread as worker thread.
        /// </summary>
        /// <param name="methodToExecute">The method to execute.</param>
        [System.Diagnostics.DebuggerStepThrough]
        public void Execute(ExecutionDelegateNoParameter methodToExecute)
        {
            IEnumerator<SwitchExecutionContext> enumerator = methodToExecute();

            if (SyncMode || (SynchronizationContext.Current != null && SynchronizationContext.Current.Equals(uiSyncContext)))
            {
                while (enumerator.MoveNext())
                {
                }
            }
            else
            {
                bool moveNext = true;
                SwitchExecutionContext mode = SwitchExecutionContext.ToUIThread;
                while (moveNext)
                {
                    if (mode == SwitchExecutionContext.ToUIThread)
                    {
                        //On UI thread
                        uiSyncContext.Send(delegate
                        {
                            moveNext = enumerator.MoveNext();
                            if (moveNext)
                            {
                                mode = enumerator.Current;
                            }
                        }, null);
                    }
                    else
                    {
                        //On the current (async) thread
                        moveNext = enumerator.MoveNext();
                        if (moveNext)
                        {
                            mode = enumerator.Current;
                        }
                    }
                }
            }

            
        }

        #region Property: bool IsBusy
        /// <summary>
        /// Holder for IsBusy
        /// </summary>
        private bool isBusy;

        /// <summary>
        /// <para>IsBusy property</para>
        /// <para>Indicates whether the executor is busy with an execution.</para>
        /// <para>Can be cound to in the UI - it fires the INotifyPropertyChanged.PropertyChanged event</para>
        /// </summary>
        public bool IsBusy
        {
            get { return isBusy; }
            private set
            {
                isBusy = value;
                FirePropertyChanged("IsBusy");
            }
        }
        #endregion

        #region INotifyPropertyChanged Members

        private void FirePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }


        public event PropertyChangedEventHandler PropertyChanged;

        #endregion
    }

    /// <summary>
    /// Delegate for a method with parameter
    /// </summary>
    public delegate IEnumerator<SwitchExecutionContext> ExecutionDelegate<InputType>(InputType parameter);

    /// <summary>
    /// delegate for method without parameter
    /// </summary>
    public delegate IEnumerator<SwitchExecutionContext> ExecutionDelegateNoParameter();

    /// <summary>
    /// Used internally to start the execution on a separate thread
    /// </summary>
    delegate void NoParametersDelegate();

    /// <summary>
    /// Enumeration for the two supported execution contexts
    /// </summary>
    public enum SwitchExecutionContext
    {
        /// <summary>
        /// Execution Context is UI Thread 
        /// </summary>
        ToUIThread, 
        /// <summary>
        /// Execution Context is Worker Thread 
        /// </summary>
        ToWorkerThread
    }
}
